import json
from typing import Any, Callable, Dict, Tuple, Union

from typeguard import check_type

from trubrics.context import ValidationContext
from trubrics.exceptions import ValidationOutputError
from trubrics.modellers.classifier import Classifier

validation_output_type = Tuple[bool, Dict[str, Union[str, int, float]]]


def validation_output(func: Callable) -> Callable:
    """Decorative function for validation point outputs."""

    def inner(*args, **kwargs) -> ValidationContext:
        output = func(*args, **kwargs)

        try:
            check_type("output", output, validation_output_type)
        except TypeError:
            raise ValidationOutputError(
                f"Each validation should return an outcome and a result with type: {validation_output_type}"
            )

        outcome, result = output

        outcome = _pass_or_fail(outcome)

        if len(args) > 1:
            args = args[1:]  # ignore self arg for class methods
        else:
            args = ()

        jsonable_args = [arg for arg in args if _is_jsonable(arg, raise_error=True)]
        jsonable_kwargs = {key: val for key, val in kwargs.items() if _is_jsonable(val, raise_error=True)}
        _is_jsonable(result, raise_error=True)

        return ValidationContext(
            validation_type=func.__name__,
            validation_kwargs={"args": jsonable_args, "kwargs": jsonable_kwargs},
            outcome=outcome,
            result=result,
        )

    return inner


def _is_jsonable(x: Any, raise_error: bool = False) -> bool:
    try:
        json.dumps(x)
        return True
    except (OverflowError, TypeError) as e:
        if isinstance(x, Classifier):
            return False
        elif raise_error:
            raise e
        else:
            return False


def _pass_or_fail(condition: bool) -> str:
    return "pass" if condition else "fail"



from typing import Optional

from trubrics.context import DataContext, ModelContext, TrubricContext
from trubrics.exceptions import UnknownValidationError
from trubrics.validators.base import Validator


def run_trubric(
    data_context: DataContext,
    model_context: ModelContext,
    trubric: TrubricContext,
    custom_validator: Optional[Validator] = None,
):
    if custom_validator is not None:
        model_validator = custom_validator
    else:
        model_validator = Validator(data=data_context, model=model_context)
    for validation in trubric.validations:
        args = validation.validation_kwargs["args"]
        kwargs = validation.validation_kwargs["kwargs"]
        try:
            result = getattr(model_validator, validation.validation_type)(*args, **kwargs)
            yield validation.validation_type, validation.severity, result.outcome
        except AttributeError:
            raise UnknownValidationError(
                f"The validation '{validation.validation_type}' does not appear to belong to a validator."
                " Try adding the object that generated the validation to the 'custom_validator' parameter."
            )



from trubrics.context import DataContext, ModelContext
from trubrics.modellers.base import Modeller


class Regressor(Modeller):
    """Classifier class with methods combining data and model contexts."""

    def __init__(self, data: DataContext, model: ModelContext):
        super().__init__(data, model)

    def explore_test_set_errors(self):
        """Filter the testing data on errors."""
        raise NotImplementedError()



import pandas as pd

from trubrics.context import DataContext, ModelContext
from trubrics.modellers.base import Modeller


class Classifier(Modeller):
    """Classifier class with methods combining data and model contexts."""

    def __init__(self, data: DataContext, model: ModelContext):
        super().__init__(data, model)

    def explore_test_set_errors(self, business_columns: bool = False) -> pd.DataFrame:
        """Filter the testing data on errors.

        Show all errors (with associated features) on the test set, returning a dataframe with
        either the raw column names or the renamed business columns.

        TODO: Fix if target column is renamed.

        Parameters
        ----------
        business_columns
            Set to *false* as a default, this flag allows you to return business column names or
            raw column names.
        """

        def _filter_errors(df):
            predict_col = f"{self.data.target_column}_predictions"
            assign_kwargs = {predict_col: self.predict()}
            return df.assign(**assign_kwargs).loc[lambda x: x[self.data.target_column] != x[predict_col], :]

        if business_columns:
            return _filter_errors(self.data.renamed_testing_data)
        else:
            return _filter_errors(self.data.testing_data)



class ValidationOutputError(BaseException):
    """An exception signalling that the output from a validation function is invalid."""


class PandasSchemaError(BaseException):
    """An exception signalling that two DataFrames do not have the same schema (column names & types)."""


class ClassifierNotSupportedError(BaseException):
    """An exception signalling that the output from a validation function is invalid."""


class UnknownEstimatorType(BaseException):
    """An exception signalling that an estimator type is not recognised."""


class UnknownValidationError(BaseException):
    """An exception signalling that validation is not recognised by a given validator."""



from abc import ABC, abstractmethod
from typing import Optional

import pandas as pd

from trubrics.context import DataContext, ModelContext


class BaseModeller(ABC):
    @abstractmethod
    def __init__(self, data: DataContext, model: ModelContext):
        self.data = data
        self.model = model

    @abstractmethod
    def predict(self, data: Optional[pd.DataFrame] = None) -> pd.Series:
        """Use the estimator to predict on the test data."""

    @abstractmethod
    def explore_test_set_errors(self):
        """Filter the testing data on errors."""

    @abstractmethod
    def compute_performance_on_test_set(self) -> float:
        """Calculate the performance on the test set with the evaluation function."""


class Modeller(BaseModeller):
    def __init__(self, data: DataContext, model: ModelContext):
        self.data = data
        self.model = model

    def predict(self, data: Optional[pd.DataFrame] = None) -> pd.Series:
        """Predict function called on model from model context."""
        try:
            if data is not None:
                return self.model.estimator.predict(data)
            return self.model.estimator.predict(self.data.testing_data[self.data.features])
        except AttributeError as error:
            raise AttributeError("Model has no .predict() method.") from error

    def explore_test_set_errors(self):
        """Filter the testing data on errors."""
        raise NotImplementedError()

    def compute_performance_on_test_set(self) -> float:
        """Calculate the performance on the test set with evaluation function."""
        predictions = self.predict()
        actuals = self.data.testing_data[self.data.target_column]
        eval_kwargs = self.model.evaluation_kwargs
        if eval_kwargs:
            return self.model.evaluation_function(predictions, actuals, **eval_kwargs)
        else:
            return self.model.evaluation_function(predictions, actuals)



"""
Something
"""

from typing import Any, Optional

from pydantic import BaseModel, validator

from trubrics.context import DataContext, ModelContext
from trubrics.validators.base import Validator


class TrubricRun(BaseModel):
    """Context for a trubrics run object."""

    data_context: DataContext
    model_context: ModelContext
    trubric_path: str
    custom_validator: Optional[Any]

    @validator("custom_validator")
    def validate_some_foo(cls, val):
        if issubclass(type(val), Validator):
            return val
        raise TypeError("Wrong type for 'custom_validator', must be subclass of Validator.")



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import requests  # type: ignore
from pydantic import BaseModel, validator

from trubrics.exceptions import PandasSchemaError
from trubrics.utils.pandas import schema_is_equal


class ModelContext(BaseModel):
    """
    The ModelContext wraps models into a trubrics friendly format.

    Note:
        The ModelContext *must contain* at least an estimator and an evaluation attribute.
        Default values are set for all other attributes.
        Currently, estimator and evaluation functions are defined with scikit-learn convention:

            * Estimators: https://scikit-learn.org/stable/developers/develop.html
            * Classification Metrics:
                https://scikit-learn.org/stable/modules/model_evaluation.html#classification-metrics
            * Regression Metrics: https://scikit-learn.org/stable/modules/model_evaluation.html#regression-metrics

    Attributes:
        name: ModelContext name. Required for trubrics UI tracking.
        version: ModelContext version. Required for trubrics UI tracking.
        estimator: An estimator (or model) that can be a classifier or regressor.
        evaluation_function: An evaluation function that takes arguments (y_true, y_pred, ...)
        evaluation_kwargs: Optional[Dict[str, Union[bool, str, int, float, None]]] = None

    """

    name: str = "my_model"
    version: float = 0.1
    estimator: Any
    evaluation_function: Any
    evaluation_kwargs: Optional[Dict[str, Union[bool, str, int, float, None]]] = None

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True
        extra = "forbid"

    @property
    def evaluation_function_name(self) -> str:
        """The scikit-learn name of the evaluation function."""
        return self.evaluation_function.__name__

    @property
    def estimator_type(self) -> str:
        """The scikit-learn type of the estimator."""
        return self.estimator._estimator_type


class DataContext(BaseModel):
    """
    The DataContext wraps data into a trubrics friendly format.

    Note:
        The DataContext *must contain* at least a testing_data and a target_column attribute.
        Default values are set for all other attributes.

    Attributes:
        name: DataContext name. Required for trubrics UI tracking.
        version: DataContext version. Required for trubrics UI tracking.
        training_data: Dataframe of the training data, that can be used in some
                       validations. E.g. data leakage validations between train and test sets.
        testing_data: Dataframe that all validations are executed against.
        categorical_columns: List of categorical names of the train & test datasets.
        business_columns: Mapping between dataset column names and comprehensible column names to be displayed to users.
        target_column: Name of target column.
    """

    name: str = "my_dataset"
    version: float = 0.1
    training_data: Optional[pd.DataFrame] = None
    testing_data: pd.DataFrame
    categorical_columns: Optional[List[str]] = None
    business_columns: Optional[Dict[str, str]] = None
    target_column: str

    class Config:
        allow_mutation = False
        arbitrary_types_allowed = True
        extra = "forbid"

    @property
    def features(self) -> List[str]:
        """Features are here defined as all testing column names excluding the target feature."""
        return [col for col in self.testing_data.columns if col != self.target_column]

    @property
    def renamed_testing_data(self) -> pd.DataFrame:
        """Renamed testing data with business columns"""
        if self.business_columns is None:
            raise TypeError("Business columns must be set to rename testing features.")
        return self.testing_data.rename(columns=self.business_columns)

    @validator("testing_data")
    def testing_and_training_must_have_same_schema(cls, v: pd.DataFrame, values: List[Any]):
        if values["training_data"] is not None and not schema_is_equal(v, values["training_data"]):  # type: ignore
            raise PandasSchemaError("Testing and training data must have identical schemas.")
        return v

    @validator("target_column")
    def target_column_must_be_in_data(cls, v: str, values: List[Any]):
        if v not in values["testing_data"].columns:  # type: ignore
            raise KeyError("Target column must be in testing_data column names.")
        return v

    @validator("categorical_columns")
    def categorical_columns_must_be_in_data(cls, v: str, values: List[Any]):
        if not set(v).issubset(values["testing_data"].columns):  # type: ignore
            raise KeyError("All categorical columns must be in testing_data column names.")
        return v

    @validator("business_columns")
    def business_columns_must_be_in_data(cls, v: str, values: List[Any]):
        if not set(v.keys()).issubset(values["testing_data"].columns):  # type: ignore
            raise KeyError("All business columns must be in testing_data column names.")
        return v

    @validator("target_column")
    def target_column_must_not_be_in_categoricals(cls, v: str, values: List[Any]):
        if values["categorical_columns"] is not None and v in values["categorical_columns"]:  # type: ignore
            raise Exception(
                "Target column should not feature as a categorical column. Categorical columns only refer to features."
            )
        return v


class FeedbackContext(BaseModel):
    """Context for feedback given by a user from a UI component."""

    feedback_type: Optional[str]
    metadata: Dict[str, Union[List[Any], str, int, float, dict]]


def _validation_context_example():
    return {
        "example": {
            "validation_type": "validate_performance_against_threshold",
            "validation_kwargs": {"args": [], "kwargs": {"threshold": 0.8}},
            "outcome": "fail",
            "severity": "error",
            "result": {"performance": "0.79"},
        }
    }


class ValidationContext(BaseModel):
    """Context for a single validation point of a model."""

    validation_type: str
    validation_kwargs: Dict[str, Optional[Any]]
    outcome: str
    severity: str = "error"
    result: Optional[Dict[str, Union[str, int, float]]]

    class Config:
        extra = "forbid"
        validate_assignment = True
        schema_extra = _validation_context_example()

    @validator("severity")
    def severity_must_be(cls, v: str):
        severity_values = ["error", "warning", "experiment"]
        if v not in severity_values:
            raise KeyError(f"Severity must be set to: {severity_values}.")
        return v

    @validator("outcome")
    def outcome_must_be(cls, v: str):
        outcome_values = ["pass", "fail"]
        if v not in outcome_values:
            raise KeyError(f"Outcome must be set to: {outcome_values}.")
        return v


class TrubricContext(BaseModel):
    """
    Context for a Trubric, or set of validation points.

    Attributes:
        name: Trubric name.
        model_context_name: model context name (from ModelContext)
        model_context_version: model context version (from ModelContext)
        data_context_name: data context name (from DataContext)
        data_context_version: data context version (from DataContext)
        metadata: free textual metadata field
        validations: list of validations (defined by ValidationContext)
    """

    name: str = "my_trubric"
    model_context_name: str
    model_context_version: float
    data_context_name: str
    data_context_version: float
    metadata: Optional[Dict[str, str]] = None
    validations: List[ValidationContext]

    class Config:
        extra = "forbid"
        schema_extra = {
            "example": {
                "name": "my_first_trubric",
                "model_context_name": "my_model",
                "model_context_version": 0.1,
                "data_context_name": "my_dataset",
                "data_context_version": 0.1,
                "metadata": {},
                "validations": [_validation_context_example()],
            }
        }

    def save_local(self, path: str):
        if path is None:
            raise Exception("Specify the local path where you would like to save your Trubric json.")
        with open(Path(path) / f"{self.name}.json", "w") as file:
            file.write(self.json())

    def save_ui(self, local_port: int):
        url = f"http://localhost:{local_port}"
        headers = {"Content-type": "application/json"}
        requests.post(
            url + "/api/trubrics/",
            data=self.json(),
            headers=headers,
        )



import logging
from typing import Optional, Tuple, Union

import pandas as pd
import streamlit as st
from pandas.api.types import is_numeric_dtype

from trubrics.context import FeedbackContext
from trubrics.exceptions import PandasSchemaError
from trubrics.modellers.classifier import Classifier
from trubrics.utils.loader import save_validation_to_json
from trubrics.utils.pandas import schema_is_equal

logger = logging.getLogger(__name__)


class StreamlitComponent(Classifier):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def generate_what_if(self, wi_data: Optional[pd.DataFrame] = None) -> pd.DataFrame:
        """
        Generate a what-if tool based on a DataFrame input.
        If no DataFrame specified, the DataContext testing data is used.

        TODO: check if total columns > N, option to select top n features based on dict of feature importance
        """
        if wi_data is None:
            wi_data = self.data.testing_data.drop(columns=[self.data.target_column])
        else:
            if not schema_is_equal(wi_data, self.data.testing_data):
                raise PandasSchemaError("Schemas of provided data and DataContext testing data are different.")
            else:
                wi_data = wi_data.drop(columns=[self.data.target_column])

        out_df = pd.DataFrame()
        for col, dtype in wi_data.dtypes.to_dict().items():
            series = wi_data[col]
            if self.data.business_columns is not None and col in self.data.business_columns.keys():
                renamed_col = self.data.business_columns[col]
            else:
                renamed_col = col
            if self.data.categorical_columns is None:
                raise ValueError("Categorical columns must be specified for the What-If tool.")
            if col in self.data.categorical_columns:
                if is_numeric_dtype(dtype.type):
                    out_df[col] = [
                        st.slider(
                            renamed_col,
                            min_value=int(series.min()),
                            max_value=int(series.max()),
                            value=round(series.mean()),
                        )
                    ]
                else:
                    out_df[col] = [st.selectbox(renamed_col, tuple(series.dropna().unique()))]
            elif is_numeric_dtype(dtype.type):
                out_df[col] = [
                    st.number_input(
                        renamed_col,
                        min_value=int(series.min()),
                        max_value=int(series.max()),
                        value=int(series.mean()),
                        step=None,
                    )
                ]
            else:
                raise NotImplementedError(f"Columns '{col}' type is not recognised.")
        return out_df

    def feedback(
        self, what_if_df: pd.DataFrame, model_prediction: Union[str, int, bool, float], tracking: bool = False
    ):
        """Get user feedback and save"""
        feedback_type: str = st.selectbox(
            "Choose feedback type:",
            (
                "Single prediction error",
                "Bias",
                "Important features",
                "Other",
            ),
        )

        metadata = {}
        what_if_input = what_if_df.to_dict("records")[0]
        if feedback_type == "Other":
            metadata["description"] = st.text_input(label="", value="Send free text feedback here")
            metadata["what_if_input"] = what_if_input
            metadata["model_prediction"] = model_prediction

        elif feedback_type == "Single prediction error":
            metadata["corrected_prediction"], metadata["description"] = self._collect_single_edge_case()
            metadata["what_if_input"] = what_if_input

        elif feedback_type == "Important features":
            (
                metadata["selected_feature"],
                metadata["top_n_feature"],
                metadata["description"],
            ) = self._collect_important_features_feedback()

        elif feedback_type == "Bias":
            metadata["description"] = "Feedback on bias."

        else:
            raise NotImplementedError()

        single_test = FeedbackContext(feedback_type=feedback_type, metadata=metadata)
        if st.button("Send feedback"):
            save_validation_to_json(trubric_context=single_test, tracking=tracking)
            logger.info(f"Predictions saved {'to Trubrics UI' if tracking else 'locally'}.")
            st.balloons()

    def _collect_single_edge_case(self) -> Tuple[Union[str, int, None], str]:
        """
        Collect correct prediction for the single edge case flag.
        """
        st.write(
            self.__feedback_type_description(
                "you are signalling that the combination of all features is a critical edge case that we must test for."
            )
        )
        corrected_prediction = st.selectbox(
            "The model prediction for this edge case should be:",
            tuple(self.data.testing_data[self.data.target_column].unique()),
        )
        description = "A single edge case."
        return corrected_prediction, description

    def _collect_important_features_feedback(self) -> Tuple[str, int, str]:
        st.write(
            self.__feedback_type_description(
                "you are signalling that a given feature must be in the top N most important features."
            )
        )
        features = self.data.features
        selected_feature = st.selectbox("Choose model feature:", (features))
        top_n_feature = st.slider(
            "The selected feature should be in the top ... features:", min_value=1, max_value=len(features)
        )
        description = "Most important features."
        return selected_feature, top_n_feature, description

    @staticmethod
    def __feedback_type_description(error_description: str) -> str:
        return f"Feedback type description: {error_description}"



from typing import Dict, Union

import numpy as np
import pandas as pd

from trubrics.context import DataContext, ModelContext
from trubrics.exceptions import ClassifierNotSupportedError, UnknownEstimatorType
from trubrics.modellers.classifier import Classifier
from trubrics.modellers.regressor import Regressor
from trubrics.validators.validation_output import (
    validation_output,
    validation_output_type,
)


class Validator:
    def __init__(self, data: DataContext, model: ModelContext):
        self.trubrics_model: Union[Classifier, Regressor]
        if model.estimator_type == "classifier":
            self.trubrics_model = Classifier(data=data, model=model)
        elif model.estimator_type == "regressor":
            self.trubrics_model = Regressor(data=data, model=model)
        else:
            UnknownEstimatorType(
                f"Estimator type {model.estimator_type} not supported. Please use a 'regressor' or 'classifier'"
                " estimator."
            )
        self.trubrics_model_type = model.estimator_type
        self.trubrics_model_eval_func = model.evaluation_function
        self.test_data = data.testing_data
        self.features = data.features
        self.target = data.target_column

    @validation_output
    def validate_single_edge_case(self, edge_case_data, desired_output):
        return self._validate_single_edge_case(edge_case_data, desired_output)

    def _validate_single_edge_case(
        self, edge_case_data: Dict[str, Union[str, int, float]], desired_output: Union[int, float]
    ) -> validation_output_type:
        """
        Single edge case validation.
        """
        prediction = self._predict_from_dict(edge_case_data=edge_case_data)
        return prediction == desired_output, {"prediction": prediction}

    @validation_output
    def validate_single_edge_case_in_range(self, edge_case_data, lower_output, upper_output):
        return self._validate_single_edge_case_in_range(edge_case_data, lower_output, upper_output)

    def _validate_single_edge_case_in_range(
        self,
        edge_case_data: Dict[str, Union[str, int, float]],
        lower_output: Union[int, float],
        upper_output: Union[int, float],
    ) -> validation_output_type:
        """
        Single edge case validation in range.
        """
        if lower_output >= upper_output:
            raise ValueError("lower_output must be strictly inferior to upper_output.")
        if self.trubrics_model_type == "classifier":
            raise ClassifierNotSupportedError("Model type 'classifier' not supported for a range output.")
        prediction = self._predict_from_dict(edge_case_data=edge_case_data)

        return prediction > lower_output and prediction < upper_output, {"prediction": prediction}

    @validation_output
    def validate_performance_against_threshold(self, threshold):
        return self._validate_performance_against_threshold(threshold)

    def _validate_performance_against_threshold(self, threshold: float) -> validation_output_type:
        """
        Compares performance of a model on a dataset to a hard coded threshold value.
        """
        performance = self.trubrics_model.compute_performance_on_test_set()

        if self.trubrics_model_eval_func.__name__ == "accuracy_score":
            return bool(performance > threshold), {"performance": performance}
        elif self.trubrics_model_eval_func.__name__ == "mean_squared_error":
            return bool(performance < threshold), {"performance": performance}
        else:
            raise NotImplementedError("The evaluation type is not recognized.")

    @validation_output
    def validate_biased_performance_across_category(self, category, threshold):
        return self._validate_biased_performance_across_category(category, threshold)

    def _validate_biased_performance_across_category(self, category: str, threshold: float) -> validation_output_type:
        """
        Calculates various performance for all values in a category and validates for
        the maximum difference in performance inferior to the threshold value.

        TODO:
        - More complex threshold function
        - Modify cardinality

        To add to output report:
        - Show distributions of category variables
        - Performance plots of results
        """
        cat_values = list(self.test_data[category].unique())
        if len(cat_values) > 20:
            raise Exception(f"Cardinality of {len(cat_values)} too high for performance test.")
        if len(cat_values) < 1:
            raise Exception(f"Category '{category}' has a single value.")
        if category not in self.test_data.columns:
            # TODO: check when categorical columns are specified
            raise KeyError(f"Column '{category}' not found in dataset.")
        result: Dict[str, Union[int, float]] = {}
        for value in cat_values:
            if value not in [np.nan, None]:
                if isinstance(value, str):
                    filtered_data = self.test_data.query(f"`{category}`=='{value}'")
                else:
                    filtered_data = self.test_data.query(f"`{category}`=={value}")
                predictions = self.trubrics_model.predict(filtered_data.loc[:, self.features])
                result[value] = float(self.trubrics_model_eval_func(filtered_data[self.target], predictions))
        max_performance_difference = max(result.values()) - min(result.values())

        return max_performance_difference < threshold, {"max_performance_difference": max_performance_difference}

    @validation_output
    def validate_feature_in_top_n_important_features(self, feature, feature_importance, top_n_features):
        return self._validate_feature_in_top_n_important_features(feature, feature_importance, top_n_features)

    @staticmethod
    def _validate_feature_in_top_n_important_features(
        feature: str, feature_importance: Dict[str, float], top_n_features: int
    ) -> validation_output_type:
        """
        Verifies that a given feature is in the top n most important features.
        """
        count = 0
        for importance in feature_importance.values():
            if importance > feature_importance[feature]:
                count += 1

        return count < top_n_features, {"feature_importance_ranking": count}

    def _predict_from_dict(self, edge_case_data: Dict[str, Union[str, int, float]]) -> Union[int, float]:
        edge_case_data_df = pd.DataFrame.from_records(edge_case_data, index=["0"], columns=self.features)
        prediction = self.trubrics_model.predict(data=edge_case_data_df)[
            0
        ].item()  # .item() converts numpy to python type, in order to be serialised to json
        return prediction
